/**
 * Title:        CoCompose
 * Copyright:    Copyright (c) 2002, 2003
 * Author:       Dennis Wagelaar
 * Organisation: University of Twente, Vrije Universiteit Brussel
 * Description:  CoCompose Design Language Tool
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License, version 2 as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  (See the file "COPYING" that is included with this source distribution.)
 */

package constraints;

import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import EDU.auburn.VGJ.graph.Node;
import EDU.utwente.CoCompose.CoCompose;
import EDU.utwente.CoCompose.lang.AbstractConcept;
import EDU.utwente.CoCompose.lang.Composite;
import EDU.utwente.CoCompose.lang.CompositeRole;
import EDU.utwente.CoCompose.lang.Concept;
import EDU.utwente.CoCompose.lang.Model;
import EDU.utwente.CoCompose.util.*;

/**
 * Constraint class for use as correctness or completeness constraint.
 * Enforces single inheritance for Java classes.
 *
 * @author Dennis Wagelaar
 * @version 0.1.18, 24 November 2003
 */

public class InheritanceConstraint implements Constraint
{
  //////////////////////////////////////////////////////////////////////////////
  // Protected methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Checks whether the given form can be interpreted as a class.
   * @param form The concept form.
   * @return True if form can be interpreted as a class.
   */
  protected boolean isClass(String form)
  {
    if (form == null) return false;
    return !((form.equals("")) || (form.endsWith("Interface")));
  } //isClass

  /**
   * Checks whether the given child concept doesn't inherit from multiple 
   * Class form concepts.
   * @param child the child concept to check
   * @param parents the parent concepts to check
   * @throws CoComposeException if child inherits from multiple Classes
   */
  protected void checkParents(Concept child, Collection parents) 
    throws CoComposeException
  {
    boolean classFound = false;
    Iterator pars = parents.iterator(); 
    while (pars.hasNext())
    {
      Object node = pars.next();
      if (node instanceof Concept)
      {
        String form = ((Concept) node).getForm();
        if (isClass(form))
        {
          if (!classFound)
          {
            classFound = true;
          }
          else
          {
            throw new CoComposeException(child + 
              " inherits from a Class more than once!", child);
          }
        }
      }
    } //while pars
  } //checkParentRoles
  
  /**
   * Retrieves all parent concepts given the child concept.
   * @param child the child concept
   * @return the Set&lt;Concept&gt; of parents
   */
  protected Set getParents(Concept child)
  {
    Model model = child.getModel();
    Set parents = new HashSet();
    for (int peer = child.firstChild(); peer != -1; peer = child.nextChild())
    {
      Node node = model.getNodeFromIndex(peer);
      if (node instanceof CompositeRole)
      {
        Composite composite = ((CompositeRole) node).getComposite();
        if (((CompositeRole) node).getName().equals("Child") &&
            composite.getType().equals("Inheritance"))
        {
          CompositeRole parent = composite.getRole("Parent");
          if (parent != null)
          {
            parents.addAll(getParents(parent));
          }
        }
      }
    } //for peer
    return parents;
  } //getParents

  /**
   * Retrieves all parent concepts given the parent role.
   * @param parent the parent role
   * @return the Set&lt;Concept&gt; of parents
   */
  protected Set getParents(CompositeRole parent)
  {
    Model model = parent.getModel();
    Set parents = new HashSet();
    for (int peer = parent.firstChild(); peer != -1; peer = parent.nextChild())
    {
      Node node = model.getNodeFromIndex(peer);
      if (node instanceof Concept)
      {
        parents.add(node);
      }
    } //for peer
    return parents;
  } //getParents

  /**
   * Checks whether this constraint holds for the given composite.
   *
   * @param composite The composite to check the constraint for.
   * @throws CoComposeException if and only if the constraint doesn't hold.
   */
  protected void checkComposite(Composite composite) throws CoComposeException
  {
    Model model = composite.getModel();
    if (model == null)
      return;
    String language = model.getTargetLanguage();
    if (!language.equals("Java"))
      if (!CoCompose.isSuperLanguage(language, "Java"))
        return;
    CompositeRole childRole = composite.getRole("Child");
    if (childRole == null)
      throw new CoComposeException("Cannot find Child role in " + composite, composite);
    for (int peer = childRole.firstChild(); peer != -1; peer = childRole.nextChild())
    {
      Node node = model.getNodeFromIndex(peer);
      if (node instanceof Concept)
      {
        Set parents = getParents((Concept) node);
        checkParents((Concept) node, parents);
      }
    } //for peer
  } //checkComposite

  //////////////////////////////////////////////////////////////////////////////
  // Public methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Checks whether this constraint holds for the given concept.
   *
   * @param concept The concept to check the constraint for.
   * @throws CoComposeException if and only if the constraint doesn't hold.
   */
  public void check(AbstractConcept concept) throws CoComposeException
  {
    if (concept instanceof Composite)
    {
      checkComposite((Composite) concept);
    }
    else
    {
      throw new CoComposeException(getClass().getName() + 
        " should be applied to Composites only", concept);
    }
  } //check
} //class InheritanceConstraint
