/**
 * Title:        CoCompose
 * Copyright:    Copyright (c) 2002
 * Author:       Dennis Wagelaar
 * Organisation: University of Twente, Vrije Universiteit Brussel
 * Description:  CoCompose Design Language Tool
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License, version 2 as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  (See the file "COPYING" that is included with this source distribution.)
 */

package implementations;

import java.util.*;
import EDU.utwente.CoCompose.lang.*;
import EDU.utwente.CoCompose.translate.*;
import EDU.utwente.CoCompose.util.*;

/**
 * Implementation generator plugin class for implementing a Returns composite
 * in ConcernJ.
 *
 * @author Dennis Wagelaar
 * @version 0.1.02, 3 September 2002
 */

public class ConcernJReturns extends ConcernJImplementationGenerator
{
  //////////////////////////////////////////////////////////////////////////////
  // Constructors
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Constructs a ConcernJReturns
   */
  public ConcernJReturns()
  {
  } //ConcernJReturns

  //////////////////////////////////////////////////////////////////////////////
  // Protected methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Applies the implementation generator to an Originator concept.
   *
   * @param originator The concept to apply the role part for.
   * @param results The set of result concepts.
   * @throws CoComposeException if generating implementation failed.
   */
  protected void applyToOriginator(Concept originator, Collection results)
    throws CoComposeException
  {
    Composite composite = getComposite();
    Iterator resultIt;
    Concept result;
    String resultName, impl, form;
    int resultCount;

    // generate method return type
    resultCount = results.size();
    if (resultCount > 1)
      throw new CoComposeException("Originator can only have one Result",
        composite);
    resultIt = results.iterator();
    if (resultIt.hasNext())
    {
      result = (Concept) resultIt.next();
      resultName = result.getName();
      form = result.getForm();
      if (form.equals("Concern") || form.equals("Primitive"))
      {
        // get current implementation
        impl = getCurrentImplementation(originator);
        // insert method result type
        impl = insertReturnType(impl, resultName);
        // add Primitive implementation tag
        if (form.equals("Primitive"))
          impl = insertMethod(impl, "<" + result.getId() + ">");
        // set implementation
        setCurrentImplementation(originator, impl);
      }
      else
        throw new CoComposeException("No Result role part defined for concept form "
          + form, result);
    }
  } //applyToOriginator

  //////////////////////////////////////////////////////////////////////////////
  // Public methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Applies the implementation generator.
   *
   * @param implementation The implementation generator to apply.
   * @param generateImpl The language specific code generation operation
   *                     visitor.
   * @throws CoComposeException if generating implementation failed.
   */
  public void apply(ImplementationGenerator implementation,
    GenerateImplementationVisitor generateImpl) throws CoComposeException
  {
    Collection results = new Vector();
    Composite composite = implementation.getComposite();
    CompositeRole originatorRole = composite.getRole("Originator");
    CompositeRole resultRole = composite.getRole("Result");
    Model model = composite.getModel();
    Iterator links;
    Object link;

    // superclass apply
    super.apply(implementation, generateImpl);
    // check roles
    if (originatorRole == null)
      throw new CoComposeException("Originator role not found in composite " +
        composite.getName(), composite);
    if (resultRole == null)
      throw new CoComposeException("Result role not found in composite " +
        composite.getName(), composite);
    // retrieve results
    links = model.getLinkedNodes(resultRole).iterator();
    while (links.hasNext())
    {
      link = links.next();
      if (link instanceof EDU.utwente.CoCompose.lang.Concept)
        results.add(link);
    }
    // apply per originator concept
    links = model.getLinkedNodes(originatorRole).iterator();
    while (links.hasNext())
    {
      link = links.next();
      if (link instanceof EDU.utwente.CoCompose.lang.Concept)
        applyToOriginator((Concept) link, results);
    } //while
  } //apply
} //ConcernJReturns
