/**
 * Title:        CoCompose
 * Copyright:    Copyright (c) 2002, 2003
 * Author:       Dennis Wagelaar
 * Organisation: University of Twente, Vrije Universiteit Brussel
 * Description:  CoCompose Design Language Tool
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License, version 2 as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  (See the file "COPYING" that is included with this source distribution.)
 */

package implementations;

import java.util.*;
import EDU.utwente.CoCompose.lang.*;
import EDU.utwente.CoCompose.translate.*;
import EDU.utwente.CoCompose.util.*;

/**
 * Implementation generator plugin class for implementing a SingleInstance
 * composite in ConcernJ.
 *
 * @author Dennis Wagelaar
 * @version 0.1.06, 5 February 2003
 */

public class ConcernJSingleInstance extends ConcernJImplementationGenerator
{
  //////////////////////////////////////////////////////////////////////////////
  // Constructors
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Constructs a ConcernJSingleInstance.
   */
  public ConcernJSingleInstance()
  {
  } //ConcernJSingleInstance

  //////////////////////////////////////////////////////////////////////////////
  // Protected methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Applies the implementation generator to an Owner concept with form Concern.
   *
   * @param owner The concept to apply the role part for.
   * @param propertyC The property concept.
	 * @param classC The class concept.
   * @throws CoComposeException if generating implementation failed.
   */
  protected void applyToOwnerConcern(Concept owner, Concept propertyC,
		Concept classC) throws CoComposeException
  {
    Composite composite = getComposite();
    Collection internals = new Vector();
    Collection conditions = new Vector();
    Collection conditionSuper = new Vector();
    Collection methods = new Vector();
    Collection methodSuper = new Vector();
    Collection inputfilters = new Vector();
    Collection outputfilters = new Vector();
    Collection filtermodules = new Vector();
    Collection variables = new Vector();
    String propertyName, className, fmodName, impl, form, condSuper, methSuper;
    Hashtable parms = new Hashtable();

    impl = getCurrentImplementation(owner);
    // generate filtermodule and superimposition declarations
    fmodName = lcfirst(composite.getName()) + "Module";
    condSuper = "self <- { ";
    methSuper = "self <- { ";
    propertyName = propertyC.getName();
    className = classC.getName();
    internals.add(propertyName + " : " + className + ";");
    variables.add("private " + className + " " + propertyName + "<" +
      propertyC.getId() + ">;");
    // add Primitive implementation tag, if necessary
    if (classC.getForm().equals("Primitive"))
    	impl = insertMethod(impl, "<" + classC.getId() + ">");
    // check inputfilters
    if (inputfilters.isEmpty())
      inputfilters = null;
    // check outputfilters
    if (outputfilters.isEmpty())
      outputfilters = null;
    // check internals
    if (internals.isEmpty())
      internals = null;
    // check conditions
    if (conditions.isEmpty())
    {
      conditions = null;
      conditionSuper = null;
    }
    else
    {
      condSuper = condSuper.substring(0, condSuper.lastIndexOf(",")) + " };";
      conditionSuper.add(condSuper);
    }
    // check methods
    if (methods.isEmpty())
    {
      methods = null;
      methodSuper = null;
    }
    else
    {
      methSuper = methSuper.substring(0, methSuper.lastIndexOf(",")) + " };";
      methodSuper.add(methSuper);
    }
    // insert variables
    if (!variables.isEmpty())
      impl = insertVariables(impl, variables);
    // insert filtermodule
    impl = insertFiltermodule(impl, createFiltermodule(fmodName, internals,
      null, conditions, methods, inputfilters, outputfilters));
    // insert superimposition
    filtermodules.add("self <- { " + fmodName + " };");
    impl = insertSuperimposition(impl, null, conditionSuper, methodSuper,
      filtermodules);
    setCurrentImplementation(owner, impl);
  } //applyToOwnerConcern

  /**
   * Applies the implementation generator to an Owner concept with form Method.
   *
   * @param owner The concept to apply the role part for.
   * @param propertyC The property concept.
	 * @param classC The class concept.
   * @throws CoComposeException if generating implementation failed.
   */
  protected void applyToOwnerMethod(Concept owner, Concept propertyC,
		Concept classC) throws CoComposeException
  {
    Composite composite = getComposite();
    Hashtable parms = new Hashtable();
    String propertyName, className, impl, form;

    impl = getCurrentImplementation(owner);
    // generate method parameters
    propertyName = propertyC.getName();
    className = classC.getName();
    parms.put(propertyName, className);
    // add Primitive implementation tag, if necessary
    if (classC.getForm().equals("Primitive"))
      impl = insertMethod(impl, "<" + classC.getId() + ">");
    // add parameters
    impl = insertParameters(impl, parms);
    setCurrentImplementation(owner, impl);
  } //applyToOwnerMethod

  /**
   * Applies the implementation generator to an Owner concept.
   *
   * @param owner The concept to apply the role part for.
   * @param propertyC The property concept.
	 * @param classC The class concept.
   * @throws CoComposeException if generating implementation failed.
   */
  protected void applyToOwner(Concept owner, Concept propertyC, Concept classC)
    throws CoComposeException
  {
    String form = owner.getForm();

    // Dispatch by owner concept form
    if (form.equals("Concern"))
      applyToOwnerConcern(owner, propertyC, classC);
    else if (form.equals("Method"))
      applyToOwnerMethod(owner, propertyC, classC);
    else
      throw new CoComposeException("No Owner role part defined for concept form "
        + form, owner);
  } //applyToOwner

  /**
   * Applies the implementation generator to a UniqueRealm.
   *
   * @param realm The unique realm.
   * @param owners The set of owner concepts.
   * @param propertyC The property concept.
	 * @param classC The class concept.
   * @throws CoComposeException if generating implementation failed.
   */
  protected void applyToRealm(UniqueRealm realm, Collection owners,
		Concept propertyC, Concept classC) throws CoComposeException
  {
    Collection methodOwners = new Vector();
    Concept owner;
    Iterator ownerIt;
    Composite composite = getComposite();
    Collection internals = new Vector();
    Collection conditions = new Vector();
    Collection conditionSuper = new Vector();
    Collection methods = new Vector();
    Collection methodSuper = new Vector();
    Collection inputfilters = new Vector();
    Collection outputfilters = new Vector();
    Collection filtermodules = new Vector();
    Collection selectors = new Vector();
    Collection variables = new Vector();
    String propertyName, className, selName, fmodName, impl, form, selector,
      condSuper, methSuper;
    int ownerConcernCount = 0;
    Hashtable parms = new Hashtable();

    // init
    selName = lcfirst(composite.getName());
    fmodName = selName + "Module";
    selector = selName + " = Set(";
    condSuper = "self <- { ";
    methSuper = selName + " <- { ";
    // divide owners in Methods and Concerns
    ownerIt = owners.iterator();
    while (ownerIt.hasNext())
    {
      owner = (Concept) ownerIt.next();
      form = owner.getForm();
      if (form.equals("Concern"))
      { // add directly to selector
        selector += owner.getName() + ", ";
        ownerConcernCount++;
      }
      else if (form.equals("Method"))
        methodOwners.add(owner);
      else
        throw new CoComposeException("No Owner role part defined for concept form "
          + form, owner);
    }
    // method owner role parts can only be applied directly to the owner
    ownerIt = methodOwners.iterator();
    while (ownerIt.hasNext())
    {
      owner = (Concept) ownerIt.next();
      applyToOwnerMethod(owner, propertyC, classC);
    }
    // stop if no owner concerns
    if (ownerConcernCount == 0)
      return;
    selector = selector.substring(0, selector.lastIndexOf(",")) + ");";
    // concern owner role parts are put in a realm concern
    impl = getRealmImplementation(realm);
    // generate filtermodule and superimposition declarations
    propertyName = propertyC.getName();
		className = classC.getName();
    internals.add(propertyName + " : " + className + ";");
    variables.add("private " + className + " " + propertyName + "<" +
      propertyC.getId() + ">;");
    // add Primitive implementation tag, if necessary
    if (classC.getForm().equals("Primitive"))
      impl = insertMethod(impl, "<" + classC.getId() + ">");
    // check inputfilters
    if (inputfilters.isEmpty())
      inputfilters = null;
    // check outputfilters
    if (outputfilters.isEmpty())
      outputfilters = null;
    // check externals
    if (internals.isEmpty())
      internals = null;
    // check conditions
    if (conditions.isEmpty())
    {
      conditions = null;
      conditionSuper = null;
    }
    else
    {
      condSuper = condSuper.substring(0, condSuper.lastIndexOf(",")) + " };";
      conditionSuper.add(condSuper);
    }
    // check methods
    if (methods.isEmpty())
    {
      methods = null;
      methodSuper = null;
    }
    else
    {
      methSuper = methSuper.substring(0, methSuper.lastIndexOf(",")) + " };";
      methodSuper.add(methSuper);
    }
    // insert variables
    if (!variables.isEmpty())
      impl = insertVariables(impl, variables);
    // insert filtermodule
    impl = insertFiltermodule(impl, createFiltermodule(fmodName, internals,
      null, conditions, methods, inputfilters, outputfilters));
    // insert superimposition
    selectors.add(selector);
    filtermodules.add(selName + " <- { " + fmodName + " };");
    impl = insertSuperimposition(impl, selectors, conditionSuper, methodSuper,
      filtermodules);
    setCurrentImplementation(realm, impl);
  } //applyToRealm

  //////////////////////////////////////////////////////////////////////////////
  // Public methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Applies the implementation generator.
   *
   * @param implementation The implementation generator to apply.
   * @param generateImpl The language specific code generation operation
   *                     visitor.
   * @throws CoComposeException if generating implementation failed.
   */
  public void apply(ImplementationGenerator implementation,
    GenerateImplementationVisitor generateImpl) throws CoComposeException
  {
    Collection owners = new Vector();
		Concept propertyC, classC;
    Composite composite = implementation.getComposite();
    CompositeRole ownerRole = composite.getRole("Owner");
    CompositeRole propertyRole = composite.getRole("Property");
    CompositeRole classRole = composite.getRole("Class");
    UniqueRealm realm = composite.getUniqueRealm();
    Model model = composite.getModel();
    Iterator links;
    Object link;

    // superclass apply
    super.apply(implementation, generateImpl);
    // check roles
    if (ownerRole == null)
      throw new CoComposeException("Owner role not found in composite " +
        composite.getName(), composite);
    if (propertyRole == null)
      throw new CoComposeException("Property role not found in composite " +
        composite.getName(), composite);
    if (classRole == null)
      throw new CoComposeException("Class role not found in composite " +
        composite.getName(), composite);
    // retrieve property and class
    links = model.getLinkedNodes(propertyRole).iterator();
    propertyC = (Concept) links.next();
    links = model.getLinkedNodes(classRole).iterator();
    classC = (Concept) links.next();
    // check whether or not to use a UniqueRealm for implementation
    if (realm == model)
    { // no special realm, so apply per owner concept
      links = model.getLinkedNodes(ownerRole).iterator();
      while (links.hasNext())
      {
        link = links.next();
        if (link instanceof EDU.utwente.CoCompose.lang.Concept)
          applyToOwner((Concept) link, propertyC, classC);
      } //while
    }
    else
    { // retrieve owners and apply for UniqueRealm
      links = model.getLinkedNodes(ownerRole).iterator();
      while (links.hasNext())
      {
        link = links.next();
        if (link instanceof EDU.utwente.CoCompose.lang.Concept)
          owners.add(link);
      } //while
      applyToRealm(realm, owners, propertyC, classC);
    }
  } //apply
} //ConcernJSingleInstance
