/**
 * Title:        CoCompose
 * Copyright:    Copyright (c) 2002, 2003
 * Author:       Dennis Wagelaar
 * Organisation: University of Twente, Vrije Universiteit Brussel
 * Description:  CoCompose Design Language Tool
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License, version 2 as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  (See the file "COPYING" that is included with this source distribution.)
 */

package implementations;

import java.util.*;
import EDU.utwente.CoCompose.lang.*;
import EDU.utwente.CoCompose.translate.*;
import EDU.utwente.CoCompose.util.CoComposeException;
import EDU.utwente.CoCompose.util.UniqueRealm;

/**
 * Introduces basic funtionality for Java implementation generators.
 *
 * @author Dennis Wagelaar
 * @version 0.1.18, 22 November 2003
 */

public abstract class JavaImplementationGenerator
  implements ImplementationGeneratorI
{
  //////////////////////////////////////////////////////////////////////////////
  // Attributes
  //////////////////////////////////////////////////////////////////////////////

  /**
   * The Implementation Pattern.
   */
  private ImplementationGenerator implGen = null;
  /**
   * The Generate Implementation operation visitor.
   */
  private GenerateImplementationVisitor genImplVis = null;

  //////////////////////////////////////////////////////////////////////////////
  // Constructors
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Constructs a JavaImplementationGenerator.
   */
  public JavaImplementationGenerator()
  {
  } //JavaImplementationGenerator

  //////////////////////////////////////////////////////////////////////////////
  // Protected methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Returns the corresponding ImplementationGenerator.
   *
   * @return The ImplementationGenerator.
   */
  protected ImplementationGenerator getImplGen()
  {
    return implGen;
  } //getImplGen

  /**
   * Returns the corresponding GenerateImplementationVisitor.
   *
   * @return The GenerateImplementationVisitor.
   */
  protected GenerateImplementationVisitor getVisitor()
  {
    return genImplVis;
  } //getVisitor

  /**
   * Retrieves current implementation for a concept.
   * @param concept The concept to get the implementation for.
   * @return The implementation.
   */
  protected String getCurrentImplementation(Concept concept)
  {
    return getVisitor().getImpl(concept);
  } //getCurrentImplementation

  /**
   * Retrieves current implementation for a unique realm.
   * @param realm The unique realm to get the implementation for.
   * @return The implementation.
   */
  protected String getCurrentImplementation(UniqueRealm realm)
  {
    return getVisitor().getImpl(realm);
  } //getCurrentImplementation

  /**
   * Stores current implementation for a concept.
   * @param concept The concept to get the implementation for.
   * @param impl The implementation.
   */
  protected void setCurrentImplementation(Concept concept, String impl)
  {
    getVisitor().setImpl(concept, impl);
  } //setCurrentImplementation

  /**
   * Stores current implementation for a unique realm.
   * @param realm The unique realm to get the implementation for.
   * @param impl The implementation.
   */
  protected void setCurrentImplementation(UniqueRealm realm, String impl)
  {
    getVisitor().setImpl(realm, impl);
  } //setCurrentImplementation

  /**
   * Returns the composite to generate implementation for.
   *
   * @return The composite.
   */
  protected Composite getComposite()
  {
    return getImplGen().getComposite();
  } //getComposite

  /**
   * Returns the model.
   *
   * @return The Model.
   */
  protected Model getModel()
  {
    return getComposite().getModel();
  } //getModel

  /**
   * Retrieves the concepts filling the given role.<BR>
   * POST: result ::= Collection&lt;Concept&gt;
   * @param role The composite role that the requested concepts fill.
   * @return The collection of parent concepts.
   */
  protected Collection getConceptsFor(CompositeRole role)
  {
    Collection concepts = new Vector();
    Object link;
    Iterator links = getModel().getLinkedNodes(role).iterator();
    while (links.hasNext())
    {
      link = links.next();
      if (link instanceof Concept)
        concepts.add(link);
    }
    return concepts;
  } //getConceptsFor

  /**
   * Changes the first character of the string to lowercase.
   *
   * @param str The string to manipulate.
   */
  protected String lcfirst(String str)
  {
    str = str.substring(0, 1).toLowerCase() + str.substring(1);
    return str;
  } //lcfirst

  /**
   * Changes the first character of the string to uppercase.
   *
   * @param str The string to manipulate.
   */
  protected String ucfirst(String str)
  {
    str = str.substring(0, 1).toUpperCase() + str.substring(1);
    return str;
  } //ucfirst

  /**
   * Returns the name of a concept adapted as instance name.
   * @param concept The concept to name.
   * @return The instance name.
   */
  protected String getInstanceName(AbstractConcept concept)
  {
    return lcfirst(concept.getName());
  } //getInstanceName

  /**
   * Checks whether the given form can be interpreted as a class.
   * @param form The concept form.
   * @return True if form can be interpreted as a class.
   */
  protected boolean isClass(String form)
  {
    return (!form.endsWith("Interface"));
  } //isClass

  /**
   * Checks whether the given form can be interpreted as an interface.
   * @param form The concept form.
   * @return True if form can be interpreted as an interface.
   */
  protected boolean isInterface(String form)
  {
    return form.endsWith("Interface");
  } //isInterface

  /**
   * Checks whether the given form can be interpreted as a primitive.
   * @param form The concept form.
   * @return True if form can be interpreted as a primitive.
   */
  protected boolean isPrimitive(String form)
  {
    return form.startsWith("Primitive");
  } //isPrimitive

  /**
   * Checks whether the given form can be interpreted as a method.
   * @param form The concept form.
   * @return True if form can be interpreted as a method.
   */
  protected boolean isMethod(String form)
  {
    return (form.endsWith("Method"));
  } //isMethod

  /**
   * Checks whether the given form is static.
   * @param form The concept form.
   * @return True if form is static.
   */
  protected boolean isStatic(String form)
  {
    return (form.startsWith("Static"));
  } //isStatic

  /**
   * Retrieves all concepts that can be interpreted as classes from the given
   * list of concepts.<BR>
   * PRE: concepts ::= Collection&lt;Concept&gt;<BR>
   * POST: result ::= Collection&lt;Concept&gt;
   * @param concepts The collection of concepts to retrieve from.
   * @return The collection of class concepts.
   * @throws CoComposeException
   */
  protected Collection getClasses(Collection concepts) throws CoComposeException
  {
    Collection classes = new Vector();
    Concept concept;
    String form;
    Iterator conceptIt = concepts.iterator();
    try
    {
      while (conceptIt.hasNext())
      {
        concept = (Concept) conceptIt.next();
        form = concept.getForm();
        if (isClass(form))
          classes.add(concept);
      } //while
    }
    catch (ClassCastException e)
    {
      throw new CoComposeException(getClass().getName() + ".getClasses(): " +
        "List of concepts contains non-Concept: " + e.getMessage(), getComposite());
    }
    return classes;
  } //getClasses

  /**
   * Retrieves all concepts that can be interpreted as interfaces from the given
   * list of concepts.<BR>
   * PRE: concepts ::= Collection&lt;Concept&gt;<BR>
   * POST: result ::= Collection&lt;Concept&gt;
   * @param concepts The collection of concepts to retrieve from.
   * @return The collection of class concepts.
   * @throws CoComposeException
   */
  protected Collection getInterfaces(Collection concepts) throws CoComposeException
  {
    Collection interfaces = new Vector();
    Concept concept;
    String form;
    Iterator conceptIt = concepts.iterator();
    try
    {
      while (conceptIt.hasNext())
      {
        concept = (Concept) conceptIt.next();
        form = concept.getForm();
        if (isInterface(form))
          interfaces.add(concept);
      } //while
    }
    catch (ClassCastException e)
    {
      throw new CoComposeException(getClass().getName() + ".getInterfaces(): " +
        "List of concepts contains non-Concept: " + e.getMessage(), getComposite());
    }
    return interfaces;
  } //getInterfaces

  /**
   * Retrieves all concepts that can be interpreted as primitives from the given
   * list of concepts.<BR>
   * PRE: concepts ::= Collection&lt;Concept&gt;<BR>
   * POST: result ::= Collection&lt;Concept&gt;
   * @param concepts The collection of concepts to retrieve from.
   * @return The collection of class concepts.
   * @throws CoComposeException
   */
  protected Collection getPrimitives(Collection concepts) throws CoComposeException
  {
    Collection primitives = new Vector();
    Concept concept;
    String form;
    Iterator conceptIt = concepts.iterator();
    try
    {
      while (conceptIt.hasNext())
      {
        concept = (Concept) conceptIt.next();
        form = concept.getForm();
        if (isPrimitive(form))
          primitives.add(concept);
      } //while
    }
    catch (ClassCastException e)
    {
      throw new CoComposeException(getClass().getName() + ".getPrimitives(): " +
        "List of concepts contains non-Concept: " + e.getMessage(), getComposite());
    }
    return primitives;
  } //getPrimitives

  /**
   * Inserts a primitive's implementation into a java implementation and
   * returns it.
   * @param impl The current implementation.
   * @param primitive The primitive concept.
   * @return The updated implementation.
   */
  protected String insertPrimitiveImplementation(String impl, Concept primitive)
  {
    int insertIndex;
    // skip package
    if ((insertIndex = impl.indexOf("package ")) != -1)
      insertIndex = impl.indexOf(';', insertIndex) + 1;
    else
      insertIndex = 0;
    impl = impl.substring(0, insertIndex) +
           "<" + primitive.getId() + ">\n" + 
           impl.substring(insertIndex);
    return impl;
  } //insertPrimitiveImplementation

  /**
   * Inserts the implementations of the given Primitive concepts in the given
   * implementation.
   *
   * @param impl The implementation to adapt.
   * @param primitives The Primitive concepts for which to insert the
   *        implementation.
   * @return The adapted implementation.
   * @throws CoComposeException if precondition not met.
   * PRE: primitives ::= Collection<Concept>
   */
  protected String insertPrimitiveImplementations(String impl,
    Collection primitives) throws CoComposeException
  {
    Concept primitive;
    Iterator prims = primitives.iterator();
    try
    {
      while (prims.hasNext())
      {
        primitive = (Concept) prims.next();
        impl = insertPrimitiveImplementation(impl, primitive);
      }
    }
    catch (ClassCastException e)
    {
      throw new CoComposeException(getClass().getName() + ".insertPrimitive" +
        "Implementations(): List of primitives contains non-Concept: " +
        e.getMessage(), getComposite());
    }
    return impl;
  } //insertPrimitveImplementations

  /**
   * Creates a comma-separated clause and returns it.<BR>
   * PRE: elements ::= Collection&lt;Concept&gt;
   * @param name The name of the clause, e.g. 'extends'.
   * @param elements The clause elements.
   * @return The clause string.
   * @throws CoComposeException
   */
  protected String createClause(String name, Collection elements)
    throws CoComposeException
  {
	String clause = new String();
    Iterator elts;
    Concept element;
    if (!elements.isEmpty())
    {
      try
      {
        clause = " " + name + " ";
        elts = elements.iterator();
        while (elts.hasNext())
        {
          element = (Concept) elts.next();
          clause += element.getName() + ", ";
        } //while
        clause = clause.substring(0, clause.lastIndexOf(","));
      } //try
      catch (ClassCastException e)
      {
        throw new CoComposeException(getClass().getName() + ".createClause(): " +
          "List of primitives contains non-Concept: " +
          e.getMessage(), getComposite());
      } //catch
    } //if
    return clause;
  } //createClause

  /**
   * Creates a Java method body from the given data.
   *
   * @param name The method name.
   * @param parms The method parameters (null = none).
   * @param returnType The method return type (null = void).
   * @param body The method body.
   * @return The created method implementation.
   */
  protected String createMethod(String name, Hashtable parms, String returnType,
    String body)
  {
    String method, parm, type;
    Enumeration ps;
    boolean first = true;

    // generate method body
    if (returnType == null)
      returnType = "void";
    method = "public " + returnType + " " + name + "(";
    if (parms != null)
    {
      ps = parms.keys();
      while (ps.hasMoreElements())
      {
        if (!first)
          method += ", ";
        parm = (String) ps.nextElement();
        type = (String) parms.get(parm);
        method += type + " " + parm;
        first = false;
      } //while
    }
    method += ") {\n";
    method += getVisitor().indent(body);
    method += "\n}";
    return method;
  } //createMethod

  /**
   * Inserts a method into a java implementation and returns it.
   *
   * @param impl The current implementation.
   * @param method The method implementation.
   * @return The updated implementation.
   */
  protected String insertMethod(String impl, String method)
  {
    int insertIndex;

    // get class or interface start and insertion point
    insertIndex = impl.lastIndexOf("}");
    // insert method into current implementation
    impl = impl.substring(0, insertIndex) +
           getVisitor().indent(method) + "\n" +
           impl.substring(insertIndex);
    return impl;
  } //insertMethod

  /**
   * Inserts an inheritance clause into the given implementation and returns it.
   *
   * @param impl The current implementation.
   * @param inheritance The inheritance clause.
   * @return The updated implementation.
   * @throws CoComposeException if illegal inheritance clause is attempted.
   */
  protected String insertInheritance(String impl, String inheritance)
    throws CoComposeException
  {
    int skipIndex, startIndex, insertIndex, ext1Index, ext2Index;

    // skip comments
    skipIndex = impl.indexOf("*/") + 1;
    // get class or interface start and insertion point
    if ((startIndex = impl.indexOf("public class", skipIndex)) == -1)
      startIndex = impl.indexOf("public interface", skipIndex);
    // cannot add another 'extends', unless same class
    if (((ext1Index = impl.indexOf("extends", startIndex)) != -1) &&
        ((ext2Index = inheritance.indexOf("extends", startIndex)) != -1))
    { // check for same class, if so return
      if (!impl.substring(ext1Index + 8, impl.indexOf(" ", ext1Index + 8)).equals(
          inheritance.substring(ext2Index + 8)))
      // multiple inheritance attempted
      throw new CoComposeException("Cannot inherit from multiple classes",
        getComposite());
    }
    // don't add two 'implements' keywords
    else if (((ext1Index = impl.indexOf("implements", startIndex)) != -1) &&
        ((ext2Index = inheritance.indexOf("implements", startIndex)) != -1))
    { // check for same class, if so return
      if (!impl.substring(ext1Index + 11, impl.indexOf(" ", ext1Index + 11)).equals(
          inheritance.substring(ext2Index + 11)))
      { // add to existing 'implements' clause
        impl = impl.substring(0, ext1Index + 11) +
          inheritance.substring(ext2Index + 11) + ", " +
          impl.substring(ext1Index + 11);
      }
    }
    else
    {
      insertIndex = impl.indexOf("{", startIndex) - 1;
      // insert and return
      impl = impl.substring(0, insertIndex) + inheritance +
        impl.substring(insertIndex);
    }
    return impl;
  } //insertInheritance

  /**
   * Inserts variables into the given implementation and returns it.
   *
   * @param impl The current implementation.
   * @param variables The inheritance clause.
   * @return The updated implementation.
   */
  protected String insertVariables(String impl, Collection variables)
  {
    int skipIndex, startIndex, insertIndex;
    Iterator varIt;
    String result;

    // skip comments
    skipIndex = impl.indexOf("*/") + 1;
    // get class or interface start and insertion point
    if ((startIndex = impl.indexOf("public class")) == -1)
      startIndex = impl.indexOf("public interface");
    insertIndex = impl.indexOf("{", startIndex) + 2;
    // insert and return
    result = impl.substring(0, insertIndex);
    varIt = variables.iterator();
    while (varIt.hasNext())
    {
      result += getVisitor().indent((String) varIt.next()) + "\n";
    }
    result += impl.substring(insertIndex);
    return result;
  } //insertInheritance

  /**
   * Inserts parameters into the method declaration and returns the updated
   * implementation.
   *
   * @param impl The current implementation.
   * @param parms The parameter declarations. Structure: <name> |-> <type>.
   * @return The updated implementation.
   */
  protected String insertParameters(String impl, Hashtable parms)
  {
    String result, parm, type;
    int startIndex, skipIndex;
    Enumeration ps;
    boolean first = true;

    // skip comments
    skipIndex = impl.indexOf("*/") + 1;
    // weave new parameters
    startIndex = impl.indexOf("(", skipIndex) + 1;
    result = impl.substring(0, startIndex);
    ps = parms.keys();
    while (ps.hasMoreElements())
    {
      if (!first)
        result += ", ";
      parm = (String) ps.nextElement();
      type = (String) parms.get(parm);
      result += type + " " + parm;
      first = false;
    }
    // if there were already more parameters, add another comma
    if (impl.charAt(startIndex) != ')')
      result += ", ";
    result += impl.substring(startIndex);
    return result;
  } //insertParameters

  /**
   * Inserts an import clause into the implementation and returns the updated
   * implementation.
   *
   * @param impl The current implementation.
   * @param libraries The set of libraries to be imported.
   * @return The updated implementation.
   */
  protected String insertImports(String impl, Collection libraries)
  {
    int insertIndex;
    Iterator libIt;
    String lib, result;

    // skip package
    if ((insertIndex = impl.indexOf("package ")) != -1)
      insertIndex = impl.indexOf(';', insertIndex) + 1;
    else
      insertIndex = 0;
//    startIndex = impl.indexOf("*/") + 1;
//    // get class or interface start and insertion point
//    if ((insertIndex = impl.indexOf("public class", startIndex)) == -1)
//      if ((insertIndex = impl.indexOf("public abstract class", startIndex)) == -1)
//        insertIndex = impl.indexOf("public interface", startIndex);
    // insert and return
    result = impl.substring(0, insertIndex);
    libIt = libraries.iterator();
    while (libIt.hasNext())
    {
      lib = (String) libIt.next();
      if (impl.indexOf("import " + lib) == -1)
        result += "import " + lib + ";\n";
    }
    result += impl.substring(insertIndex);
    return result;
  } //insertImports

  /**
   * Inserts a return type clause into the implementation and returns the updated
   * implementation.
   *
   * @param impl The current implementation.
   * @param type The return type.
   * @return The updated implementation.
   */
  protected String insertReturnType(String impl, String type)
  {
    int startIndex, endIndex;

    // skip comments
    startIndex = impl.indexOf("*/") + 1;
    // get insertion point (delete previous return type (void?))
    startIndex = impl.indexOf("void ", startIndex);
    endIndex = startIndex + 4;
    // insert and return
    impl = impl.substring(0, startIndex) + type + impl.substring(endIndex);
    return impl;
  } //insertReturnType

  /**
   * Inserts a method override id into the implementation and returns the
   * updated implementation.
   *
   * @param impl The current implementation.
   * @param id The overridden method id.
   * @return The updated implementation.
   */
  protected String insertMethodOverrideID(String impl, int id)
  {
    int insertIndex, startIndex;

    // skip comments
    startIndex = impl.indexOf("*/") + 1;
    // get insertion point (delete previous header)
    insertIndex = impl.indexOf("{", startIndex) - 1;
    startIndex = impl.lastIndexOf(' ', impl.lastIndexOf('(', insertIndex));
    // check if return type present
    if (impl.lastIndexOf(' ', startIndex - 1) != -1)
      startIndex = impl.lastIndexOf(' ', startIndex - 1);
    // insert and return
    impl = impl.substring(0, startIndex) + " <" + id + ".header>" + impl.substring(insertIndex);
    return impl;
  } //insertMethodOverrideID

  //////////////////////////////////////////////////////////////////////////////
  // Public methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Applies the implementation generator.
   *
   * @param implementation The implementation generator to apply.
   * @param generateImpl The language specific code generation operation
   *                     visitor.
   * @throws CoComposeException if generating implementation failed.
   */
  public void apply(ImplementationGenerator implementation,
    GenerateImplementationVisitor generateImpl) throws CoComposeException
  {
    implGen = implementation;
    genImplVis = generateImpl;
  } //apply
} //class JavaImplementationGenerator
