/**
 * Title:        CoCompose
 * Copyright:    Copyright (c) 2002, 2003
 * Author:       Dennis Wagelaar
 * Organisation: University of Twente, Vrije Universiteit Brussel
 * Description:  CoCompose Design Language Tool
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License, version 2 as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  (See the file "COPYING" that is included with this source distribution.)
 */

package implementations;

import java.util.*;
import EDU.utwente.CoCompose.translate.GenerateImplementationVisitor;
import EDU.utwente.CoCompose.translate.ImplementationGenerator;
import EDU.utwente.CoCompose.util.CoComposeException;
import EDU.utwente.CoCompose.lang.Concept;
import EDU.utwente.CoCompose.lang.Composite;
import EDU.utwente.CoCompose.lang.CompositeRole;

/**
 * Implementation generator plugin class for implementing an Inheritance composite
 * in Java.
 *
 * @author Dennis Wagelaar
 * @version 0.1.18, 22 November 2003
 */

public class JavaInheritance extends JavaImplementationGenerator
{
  //////////////////////////////////////////////////////////////////////////////
  // Protected methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Retrieves the parent role of the composite.
   * @return The parent CompositeRole.
   */
  protected CompositeRole getParentRole()
  {
    return getComposite().getRole("Parent");
  } //getParentRole

  /**
   * Retrieves the child role of the composite.
   * @return The child CompositeRole.
   */
  protected CompositeRole getChildRole()
  {
    return getComposite().getRole("Child");
  } //getChildRole

  /**
   * Checks the composite roles for existence.
   * @throws CoComposeException if a required role does not exist.
   */
  protected void checkRoles() throws CoComposeException
  {
    Composite composite = getComposite();
    if (getParentRole() == null)
      throw new CoComposeException("Parent role not found in composite " +
        composite.getName(), composite);
    if (getChildRole() == null)
      throw new CoComposeException("Child role not found in composite " +
        composite.getName(), composite);
  } //checkRoles

  /**
   * Applies the implementation generator for each Child concept.
   * @throws CoComposeException if generating implementation failed.
   */
  protected void applyPerChild() throws CoComposeException
  {
    Object link;
    Collection parents = getConceptsFor(getParentRole());
    Iterator links = getModel().getLinkedNodes(getChildRole()).iterator();
    while (links.hasNext())
    {
      link = links.next();
      if (link instanceof Concept)
        applyToChild((Concept) link, parents);
    } //while
  } //applyPerChild

  /**
   * Applies the implementation generator to a Child concept.
   * @param child The concept to apply the role part for.
   * @param parents The set of parent concepts.
   * @throws CoComposeException if generating implementation failed.
   */
  protected void applyToChild(Concept child, Collection parents)
    throws CoComposeException
  {
    // Dispatch child concept form
    if (isInterface(child.getForm()))
      applyToChildInterface(child, parents);
    else
      applyToChildClass(child, parents);
  } //applyToChild

  /**
   * Applies the implementation generator to a Child concept as a Class.
   *
   * @param child The concept to apply the role part for.
   * @param parents The set of parent concepts.
   * @throws CoComposeException if generating implementation failed.
   */
  protected void applyToChildClass(Concept child, Collection parents)
    throws CoComposeException
  {
    String impl, exts, impls;

    impl = getCurrentImplementation(child);
    // check java inheritance constraints
    if (getClasses(parents).size() > 1)
      throw new CoComposeException("Cannot inherit from multiple classes",
        getComposite());
    impl = insertPrimitiveImplementations(impl, getPrimitives(parents));
    exts = createClause("extends", getClasses(parents));
    impls = createClause("implements", getInterfaces(parents));
    impl = insertInheritance(impl, exts + impls);
    setCurrentImplementation(child, impl);
  } //applyToChildClass

  /**
   * Applies the implementation generator to a Child concept as an Interface.
   *
   * @param child The concept to apply the role part for.
   * @param parents The set of parent concepts.
   * @throws CoComposeException if generating implementation failed.
   */
  protected void applyToChildInterface(Concept child, Collection parents)
    throws CoComposeException
  {
    String impl, exts;

    impl = getCurrentImplementation(child);
    impl = insertPrimitiveImplementations(impl, getPrimitives(parents));
    exts = createClause("extends", parents);
    impl = insertInheritance(impl, exts);
    setCurrentImplementation(child, impl);
  } //applyToChildInterface

  //////////////////////////////////////////////////////////////////////////////
  // Public methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Applies the implementation generator.
   *
   * @param implementation The implementation generator to apply.
   * @param generateImpl The language specific code generation operation
   *                     visitor.
   * @throws CoComposeException if generating implementation failed.
   */
  public void apply(ImplementationGenerator implementation,
    GenerateImplementationVisitor generateImpl) throws CoComposeException
  {
    super.apply(implementation, generateImpl);
    checkRoles();
    applyPerChild();
  } //apply
} //class JavaInheritance
