/**
 * Title:        CoCompose
 * Copyright:    Copyright (c) 2002, 2003
 * Author:       Dennis Wagelaar
 * Organisation: University of Twente, Vrije Universiteit Brussel
 * Description:  CoCompose Design Language Tool
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License, version 2 as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  (See the file "COPYING" that is included with this source distribution.)
 */

package implementations;

import java.util.*;
import EDU.utwente.CoCompose.lang.*;
import EDU.utwente.CoCompose.translate.*;
import EDU.utwente.CoCompose.util.*;

/**
 * Implementation generator plugin class for implementing a Overrides composite
 * in Java.
 *
 * @author Dennis Wagelaar
 * @version 0.1.06, 4 February 2003
 */

public class JavaOverrides extends JavaImplementationGenerator
{
  //////////////////////////////////////////////////////////////////////////////
  // Constructors
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Constructs a JavaOverrides
   */
  public JavaOverrides()
  {
  } //JavaOverrides

  //////////////////////////////////////////////////////////////////////////////
  // Protected methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Retrieves the OldMethod role of the composite.
   * @return The OldMethod CompositeRole.
   */
  protected CompositeRole getOldMethodRole()
  {
    return getComposite().getRole("OldMethod");
  } //getOldMethodRole

  /**
   * Retrieves the NewMethod role of the composite.
   * @return The NewMethod CompositeRole.
   */
  protected CompositeRole getNewMethodRole()
  {
    return getComposite().getRole("NewMethod");
  } //getNewMethodRole

  /**
   * Checks the composite roles for existence.
   * @throws CoComposeException if a required role does not exist.
   */
  protected void checkRoles() throws CoComposeException
  {
    Composite composite = getComposite();
    if (getOldMethodRole() == null)
      throw new CoComposeException("OldMethod role not found in composite " +
        composite.getName(), composite);
    if (getNewMethodRole() == null)
      throw new CoComposeException("NewMethod role not found in composite " +
        composite.getName(), composite);
  } //checkRoles

  /**
   * Checks whether the amount of OldMethod concepts doesn't exceed 1.<BR>
   * PRE: oldMethods ::= Collectio&lt;Concept&gt;
   * @param oldMethods The collection of OldMethod concepts.
   * @throws CoComposeException if the amount exceeds 1.
   */
  protected void checkOldMethodsCount(Collection oldMethods)
    throws CoComposeException
  {
    if (oldMethods.size() > 1)
      throw new CoComposeException("NewMethod can only have one OldMethod",
        getComposite());
  } //checkOldMethodsCount

  /**
   * Applies the implementation generator for each NewMethod concept.
   * @throws CoComposeException if generating implementation failed.
   */
  protected void applyPerNewMethod() throws CoComposeException
  {
    Object link;
    Collection oldMethods = getConceptsFor(getOldMethodRole());
    Iterator links = getModel().getLinkedNodes(getNewMethodRole()).iterator();
    while (links.hasNext())
    {
      link = links.next();
      if (link instanceof Concept)
        applyToNewMethod((Concept) link, oldMethods);
    } //while
  } //applyPerNewMethod

  /**
   * Applies the implementation generator to an NewMethod concept.
   *
   * @param newMethod The concept to apply the role part for.
   * @param oldMethods The set of oldMethod concepts.
   * @throws CoComposeException if generating implementation failed.
   */
  protected void applyToNewMethod(Concept newMethod, Collection oldMethods)
    throws CoComposeException
  {
    Iterator oldMethodIt;
    Concept oldMethod;
    String impl;

    checkOldMethodsCount(oldMethods);
    oldMethodIt = oldMethods.iterator();
    if (oldMethodIt.hasNext())
    {
      oldMethod = (Concept) oldMethodIt.next();
      impl = getCurrentImplementation(newMethod);
      impl = insertMethodOverrideID(impl, oldMethod.getId());
      setCurrentImplementation(newMethod, impl);
    }
  } //applyToNewMethod

  //////////////////////////////////////////////////////////////////////////////
  // Public methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Applies the implementation generator.
   *
   * @param implementation The implementation generator to apply.
   * @param generateImpl The language specific code generation operation
   *                     visitor.
   * @throws CoComposeException if generating implementation failed.
   */
  public void apply(ImplementationGenerator implementation,
    GenerateImplementationVisitor generateImpl) throws CoComposeException
  {
    super.apply(implementation, generateImpl);
    checkRoles();
    applyPerNewMethod();
  } //apply
} //JavaOverrides
