/**
 * Title:        CoCompose
 * Copyright:    Copyright (c) 2002, 2003
 * Author:       Dennis Wagelaar
 * Organisation: University of Twente, Vrije Universiteit Brussel
 * Description:  CoCompose Design Language Tool
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License, version 2 as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  (See the file "COPYING" that is included with this source distribution.)
 */

package implementations;

import java.util.*;
import EDU.utwente.CoCompose.lang.*;
import EDU.utwente.CoCompose.translate.*;
import EDU.utwente.CoCompose.util.*;

/**
 * Implementation generator plugin class for implementing a SingleInstance
 * composite in Java.
 *
 * @author Dennis Wagelaar
 * @version 0.1.17, 18 November 2003
 */

public class JavaSingleInstance extends JavaImplementationGenerator
{
  //////////////////////////////////////////////////////////////////////////////
  // Constructors
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Constructs a JavaSingleInstance.
   */
  public JavaSingleInstance()
  {
  } //JavaSingleInstance

  //////////////////////////////////////////////////////////////////////////////
  // Protected methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Retrieves the Owner role of the composite.
   * @return The Owner CompositeRole.
   */
  protected CompositeRole getOwnerRole()
  {
    return getComposite().getRole("Owner");
  } //getOwnerRole

  /**
   * Retrieves the Property role of the composite.
   * @return The Property CompositeRole.
   */
  protected CompositeRole getPropertyRole()
  {
    return getComposite().getRole("Property");
  } //getPropertyRole

  /**
   * Retrieves the Class role of the composite.
   * @return The Class CompositeRole.
   */
  protected CompositeRole getClassRole()
  {
    return getComposite().getRole("Class");
  } //getClassRole

  /**
   * Checks the composite roles for existence.
   * @throws CoComposeException if a required role does not exist.
   */
  protected void checkRoles() throws CoComposeException
  {
    Composite composite = getComposite();
    if (getOwnerRole() == null)
      throw new CoComposeException("Owner role not found in composite " +
        composite.getName(), composite);
    if (getPropertyRole() == null)
      throw new CoComposeException("Property role not found in composite " +
        composite.getName(), composite);
    if (getClassRole() == null)
      throw new CoComposeException("Class role not found in composite " +
        composite.getName(), composite);
  } //checkRoles

  /**
   * Applies the implementation generator for each Owner concept.
   * @throws CoComposeException if generating implementation failed.
   */
  protected void applyPerOwner() throws CoComposeException
  {
    Object link;
    Iterator links = getModel().getLinkedNodes(getOwnerRole()).iterator();
    while (links.hasNext())
    {
      link = links.next();
      if (link instanceof Concept)
        applyToOwner((Concept) link);
    } //while
  } //applyPerOwner

  /**
   * Applies the implementation generator to an Owner concept with form Class.
   *
   * @param owner The concept to apply the role part for.
   * @param propertyC The property concept.
   * @param classC The class concept.
   * @throws CoComposeException if generating implementation failed.
   */
  protected void applyToOwnerClass(Concept owner, Concept propertyC,
		Concept classC) throws CoComposeException
  {
    Collection variables = new Vector();
    String impl, variable;

    impl = getCurrentImplementation(owner);
    // add Primitive implementation tag, if necessary
    if (isPrimitive(classC.getForm()))
      impl = insertPrimitiveImplementation(impl, classC);
    // generate variables
    variable = "private ";
    if (isStatic(propertyC.getForm()))
      variable += "static ";
    variable += classC.getName() + " " + propertyC.getName() + "<" +
      propertyC.getId() + ">;";
    variables.add(variable);
    if (!variables.isEmpty())
      impl = insertVariables(impl, variables);
    setCurrentImplementation(owner, impl);
  } //applyToOwnerClass

  /**
   * Applies the implementation generator to an Owner concept with form Method.
   *
   * @param owner The concept to apply the role part for.
   * @param propertyC The property concept.
   * @param classC The class concept.
   * @throws CoComposeException if generating implementation failed.
   */
  protected void applyToOwnerMethod(Concept owner, Concept propertyC,
		Concept classC) throws CoComposeException
  {
    Hashtable parms = new Hashtable();
    String impl;

    impl = getCurrentImplementation(owner);
    // add Primitive implementation tag, if necessary
    if (isPrimitive(classC.getForm()))
    	impl = insertPrimitiveImplementation(impl, classC);
    // generate parameters
    parms.put(propertyC.getName(), classC.getName());
    impl = insertParameters(impl, parms);
    setCurrentImplementation(owner, impl);
  } //applyToOwnerMethod

  /**
   * Applies the implementation generator to an Owner concept.
   * @param owner The concept to apply the role part for.
   * @throws CoComposeException if generating implementation failed.
   */
  protected void applyToOwner(Concept owner)
    throws CoComposeException
  {
    Concept propertyC = (Concept) getConceptsFor(getPropertyRole()).iterator().next();
    Concept classC = (Concept) getConceptsFor(getClassRole()).iterator().next();
    // Dispatch by owner concept form
    if (isMethod(owner.getForm()))
      applyToOwnerMethod(owner, propertyC, classC);
    else
      applyToOwnerClass(owner, propertyC, classC);
  } //applyToOwner

  //////////////////////////////////////////////////////////////////////////////
  // Public methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Applies the implementation generator.
   *
   * @param implementation The implementation generator to apply.
   * @param generateImpl The language specific code generation operation
   *                     visitor.
   * @throws CoComposeException if generating implementation failed.
   */
  public void apply(ImplementationGenerator implementation,
    GenerateImplementationVisitor generateImpl) throws CoComposeException
  {
    super.apply(implementation, generateImpl);
    checkRoles();
    applyPerOwner();
  } //apply
} //JavaSingleInstance
