/**
 * Title:        CoCompose
 * Copyright:    Copyright (c) 2002, 2003
 * Author:       Dennis Wagelaar
 * Organisation: University of Twente, Vrije Universiteit Brussel
 * Description:  CoCompose Design Language Tool
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License, version 2 as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  (See the file "COPYING" that is included with this source distribution.)
 */

package translate;

import java.util.*;
import EDU.utwente.CoCompose.lang.*;
import EDU.utwente.CoCompose.util.*;

/**
 * Generates ConcernJ implementation for a CoCompose model.
 *
 * @author Dennis Wagelaar
 * @version 0.1.10, 22 July 2003
 */

public class ConcernJImplementationVisitor extends JavaImplementationVisitor
{
  //////////////////////////////////////////////////////////////////////////////
  // Constants
  //////////////////////////////////////////////////////////////////////////////

  private static final String CJ_CONCERN   = "Concern";
  private static final String CJ_METHOD    = "Method";
  private static final String CJ_CONDITION = "Condition";
  private static final String CJ_FILTERM   = "Filtermodule";

  //////////////////////////////////////////////////////////////////////////////
  // Constructors
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Constructs a ConcernJImplementationVisitor.
   */
  public ConcernJImplementationVisitor()
  {
  } //ConcernJImplementationVisitor

  //////////////////////////////////////////////////////////////////////////////
  // Protected methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Implements the concept as a Concern.
   *
   * @param concept The concept to implement.
   */
  protected void conceptAsConcern(Concept concept)
  {
    String impl;

    impl = "concern " + concept.getName() + " begin\n" +
           "  implementation in java file \"" + concept.getName() +
           ".jwf\";\n" +
           "end concern " + concept.getName() + ";\n" +
           "<JAVA>\n" + // separator to create .cf and .jwf files
           "public class " + concept.getName() + " {\n" +
              getImplementation(concept, "Concern", true) + "}";
    setImpl(concept, impl);
    if (log != null)
      log.log("Implemented " + concept.getName() + " as Concern");
  } //conceptAsConcern

  /**
   * Implements the concept as a Method.
   *
   * @param concept The concept to implement.
   */
  protected void conceptAsMethod(Concept concept)
  {
    String impl;

    impl = concept.getName() + "();\n" +
           "<JAVA>\n" + // separator to create .cf and .jwf files
           "public void " + concept.getName() + "() {\n" +
           getImplementation(concept, "Method", true) + "}";
    setImpl(concept, impl);
    if (log != null)
      log.log("Implemented " + concept.getName() + " as Method");
  } //conceptAsMethod

  /**
   * Implements the concept as a Condition.
   *
   * @param concept The concept to implement.
   */
  protected void conceptAsCondition(Concept concept)
  {
    String impl;

    impl = concept.getName() + "();\n" +
           "<JAVA>\n" + // separator to create .cf and .jwf files
           "public boolean " + concept.getName() + "() {\n" +
           getImplementation(concept, "Condition", true) + "}";
    setImpl(concept, impl);
    if (log != null)
      log.log("Implemented " + concept.getName() + " as Condition");
  } //conceptAsCondition

  /**
   * Implements the concept as a Filtermodule.
   *
   * @param concept The concept to implement.
   */
  protected void conceptAsFiltermodule(Concept concept)
  {
    String impl;

    impl = "filtermodule " + concept.getName() + " begin\n" +
           getImplementation(concept, "Filtermodule", true) +
           "end filtermodule " + concept.getName() + ";";
    setImpl(concept, impl);
    if (log != null)
      log.log("Implemented " + concept.getName() + " as Filtermodule");
  } //conceptAsFiltermodule

  /**
   * Retrieves the right part from the implementation.
   *
   * @param attr The pointer attributes, if any.
   * @param impl The current implementation.
   * @return The correct part of the implementation.
   * @throws CoComposeException if the part could not be found.
   */
  protected String getImplPart(String attr, String impl)
    throws CoComposeException
  {
    int startIndex;

    if (attr != null)
    {
      if (log != null)
        log.log("Pointer attribute found");
      return getPointerAttribute(attr, impl);
    }
    else
    { // return implementation after
      // <JAVA> tag
      if ((startIndex = impl.indexOf("<JAVA>\n")) != -1)
        impl = impl.substring(startIndex + 7);
      return impl;
    }
  } //getImplPart

  /**
   * Gets the pointer attribute part of the implementation for the given
   * entity.
   *
   * @param attr The pointer attribute identifier.
   * @param impl The implementation to get the attribute part for.
   * @throws CoComposeException if the part could not be found.
   */
  protected String getPointerAttribute(String attr, String impl)
    throws CoComposeException
  {
    int startIndex, endIndex, abstrIndex;

    // Dispatch attribute
    if (attr.equals("cf_header"))
    { // get ConcernJ method header
      if (log != null)
        log.log("ConcernJ header attribute found");
      try
      {
        if ((endIndex = impl.indexOf("\n<JAVA>\n")) == -1)
          if ((endIndex = impl.indexOf("\n  <JAVA>\n")) == -1)
            endIndex = impl.indexOf("\n    <JAVA>\n");
        if (endIndex != -1)
	      impl = impl.substring(0, endIndex);
        else
          System.out.println("Warning: no <JAVA> tag found in " + impl);
      }
      catch (StringIndexOutOfBoundsException se)
      { // if this happens, there is a bug
        if (log != null)
          log.log("Error in getPointerAttribute: " + se.getMessage() +
            ", impl = \"" + impl + "\"");
        se.printStackTrace();
        throw new CoComposeException("Error in getPointerAttribute: " +
          se.getMessage() + ", impl = \"" + impl + "\"");
      } //catch
    }
    else if (attr.equals("cf_name"))
    { // get ConcernJ filter, method or concern name
      if (log != null)
        log.log("ConcernJ name attribute found");
      try
      {
        if ((startIndex = impl.indexOf("concern ")) != -1)
        { // get concern name
          startIndex += 8;
          endIndex = impl.indexOf(" ", startIndex);
          impl = impl.substring(startIndex, endIndex);
        }
        else if ((startIndex = impl.indexOf("filtermodule ")) != -1)
        { // get filtermodule name
          startIndex += 13;
          endIndex = impl.indexOf(" ", startIndex);
          impl = impl.substring(startIndex, endIndex);
        }
        else
        { // get method/condition name
          endIndex = impl.indexOf("(");
          impl = impl.substring(0, endIndex);
        }
      }
      catch (StringIndexOutOfBoundsException se)
      { // if this happens, there is a bug
        if (log != null)
          log.log("Error in getPointerAttribute: " + se.getMessage() +
            ", impl = \"" + impl + "\"");
        se.printStackTrace();
        throw new CoComposeException("Error in getPointerAttribute: " +
          se.getMessage() + ", impl = \"" + impl + "\"");
      } //catch
    }
    else
    { // Inherited attributes
      impl = super.getPointerAttribute(attr, impl);
    }
    return impl;
  } //getPointerAttribute

  /**
   * Moves the import declarations in the implementations to the right place.
   */
  protected void moveImports()
  {
    Hashtable implementations = getImpls();
    Enumeration entities = implementations.keys();
    Object entity;
    String impl, imports;
    int startIndex, importIndex, endIndex, count;

    if (log != null)
    {
      log.log("Moving import declarations");
      log.incLogLevel();
    }
    try
    {
      // move import declarations within Java part
      super.moveImports();
      // move import declarations within CF part
      while (entities.hasMoreElements())
      {
        entity = entities.nextElement();
        impl = (String) implementations.get(entity);
        // find end of concern and do reverse search
        if ((endIndex = impl.lastIndexOf("end concern")) == -1)
          continue;
        // find import declarations and move them
        count = 0;
        imports = new String();
        while ((importIndex = impl.lastIndexOf("import", endIndex)) != -1)
        { // find and copy import declaration
          endIndex = impl.indexOf(";", importIndex) + 1;
          imports += impl.substring(importIndex, endIndex) + '\n';
          // remove import declaration
          impl = impl.substring(0, importIndex).trim() + impl.substring(endIndex);
          count++;
        } //while importIndex
        // find class or interface start
        startIndex = impl.indexOf("public ");
        if (count > 0)
        { // insert import declarations and set updated implementation
          impl = impl.substring(0, startIndex) + imports +
            impl.substring(startIndex);
          if (log != null)
            log.log("Moved " + count + " import declarations " +
              entity.getClass().getName());
          implementations.put(entity, impl);
        }
      } //while impls
    }
    finally
    {
      if (log != null)
        log.decLogLevel();
    }
  } //moveImports

  /**
   * Writes the implementation of the model.
   *
   * @param model The model to write the implementation for.
   * @param outdir The directory to save the implementation files.
   * @throws CoComposeException if writing the implementation failed.
   */
  protected void writeImplementation(Model model, String outdir)
    throws CoComposeException
  {
    Hashtable implementations = getImpls();
    Enumeration entities = implementations.keys();
    Object entity;
    String impl, cf_impl, j_impl, name;
    int javaIndex;

    if (log != null)
    {
      log.log("Writing implementations to files");
      log.incLogLevel();
    }
    try
    {
      while (entities.hasMoreElements())
      {
        entity = entities.nextElement();
        if (entity instanceof EDU.utwente.CoCompose.lang.Concept)
        { // only write concerns to files
          if (!((Concept) entity).getForm().equals("Concern"))
            continue;
          name = ((Concept) entity).getName();
        }
        else if (entity instanceof EDU.utwente.CoCompose.util.UniqueRealm)
          name = ((UniqueRealm) entity).getName();
        else
          continue;
        // write implementation
        impl = (String) implementations.get(entity);
        // split .cf and .jwf file
        javaIndex = impl.indexOf("\n<JAVA>\n");
        cf_impl = impl.substring(0, javaIndex);
        j_impl = impl.substring(javaIndex + 8);
        writeFile(outdir, name + ".cf", cf_impl);
        writeFile(outdir, name + ".jwf", j_impl);
        if (log != null)
          log.log(name + " written");
      } //while
    }
    finally
    {
      log.decLogLevel();
    }
  } //writeImplementation

  //////////////////////////////////////////////////////////////////////////////
  // Public methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Returns the target language this GenerateImplementationVisitor supports.
   *
   * @return The target language.
   */
  public String getTargetLanguage()
  {
    return "ConcernJ";
  } //getTargetLanguage

  /**
   * Generates implementation for a Concept object.
   *
   * @param concept The concept to visit.
   * @throws CoComposeException if generating implementation failed.
   */
  public void visitConcept(Concept concept) throws CoComposeException
  {
    String form = concept.getForm();

    if (log != null)
    {
      log.log("Generating implementation for concept " + concept.getName());
      log.incLogLevel();
    }
    try
    {
      // now dispatch to specific code generators
      if (form.equals(CJ_CONCERN))
        conceptAsConcern(concept);
      else if (form.equals(CJ_METHOD))
        conceptAsMethod(concept);
      else if (form.equals(CJ_CONDITION))
        conceptAsCondition(concept);
      else if (form.equals(CJ_FILTERM))
        conceptAsFiltermodule(concept);
      else
        super.visitConcept(concept);
    }
    finally
    {
      if (log != null)
        log.decLogLevel();
    }
  } //visitConcept
} //class ConcernJImplementationVisitor
