/**
 * Title:        CoCompose
 * Copyright:    Copyright (c) 2002, 2003
 * Author:       Dennis Wagelaar
 * Organisation: University of Twente, Vrije Universiteit Brussel
 * Description:  CoCompose Design Language Tool
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License, version 2 as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  (See the file "COPYING" that is included with this source distribution.)
 */

package translate;

import java.io.*;
import java.util.*;
import EDU.utwente.CoCompose.lang.*;
import EDU.utwente.CoCompose.translate.*;
import EDU.utwente.CoCompose.util.*;

/**
 * Generates Java implementation for a CoCompose model.
 *
 * @author Dennis Wagelaar
 * @version 0.1.18, 22 November 2003
 */

public class JavaImplementationVisitor extends GenerateImplementationVisitor
{
  //////////////////////////////////////////////////////////////////////////////
  // Constants
  //////////////////////////////////////////////////////////////////////////////

  private static final String J_CLASS     = "Class";
  private static final String J_ABSCLASS  = "AbstractClass";
  private static final String J_INTERFACE = "Interface";
  private static final String J_PUBCONSTR = "PublicConstructorMethod";
  private static final String J_PRICONSTR = "PrivateConstructorMethod";
  private static final String J_PUBMETHOD = "PublicMethod";
  private static final String J_PRIMETHOD = "PrivateMethod";
  private static final String J_ABSMETHOD = "AbstractMethod";
  private static final String J_STAMETHOD = "StaticMethod";

  //////////////////////////////////////////////////////////////////////////////
  // Constructors
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Constructs a JavaImplementationVisitor.
   */
  public JavaImplementationVisitor()
  {
  } //JavaImplementationVisitor

  //////////////////////////////////////////////////////////////////////////////
  // Private methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Substitutes predefined variables. Known variables: <?name?>
   * @param entity The entity for which to substitute variables.
   * @param impl The entity's implementation.
   * @return The adapted implementation.
   * @throws CoComposeException if substituting failed.
   */
  private String substVars(Object entity, String impl)
    throws CoComposeException
  {
    int varIndex = -1;

    if (log != null)
      log.log("Substituting variables for " + entity);
    // search and replace variables
    while ((varIndex = impl.indexOf("<?name?>")) != -1)
    { // <?name?> variable found
      if (entity instanceof EDU.utwente.CoCompose.lang.AbstractConcept)
      {
        if (log != null)
          log.log("Substituting <?name?> variable for " + ((Concept) entity).getName());
        impl = impl.substring(0, varIndex) + ((Concept) entity).getName() +
          impl.substring(varIndex + 8);
      }
      else
      {
        if (log != null)
          log.log("Error: <?name?> variable only valid for AbstractConcept");
        throw new CoComposeException("<?name?> variable only valid for AbstractConcept");
      }
    }
    return impl;
  } //substVars
  
  /**
   * Returns pointers to any embedded Primitive concepts in the implementation
   * @param impl - the implementation to check
   * @return the Primitive pointer for impl
   */
  private String getEmbeddedPrimitives(Implementation impl)
  {
    String result = new String();
    Enumeration concepts = impl.getConstraints().keys();
    while (concepts.hasMoreElements())
    {
      AbstractConcept concept = (AbstractConcept) concepts.nextElement();
      if (concept instanceof Concept)
        if (((Concept) concept).getForm().startsWith("Primitive"))
          result += "\n<" + concept.getId() + ">";
    }
    return result;
  }
  
  //////////////////////////////////////////////////////////////////////////////
  // Protected methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Checks and returns the implementation content, if check ok.
   * @param impl - the implementation to check
   * @return the content of impl, if check ok
   * @throws CoComposeException if check not ok
   */
  protected String getImplementation(Implementation impl) throws CoComposeException
  {
    String result = super.getImplementation(impl);
    result += getEmbeddedPrimitives(impl);
    return result;
  } //getImplementation

  /**
   * Implements the concept as a Class.
   *
   * @param concept The concept to implement.
   * @param form The exact concept form.
   */
  protected void conceptAsClass(Concept concept, String form)
  {
    String impl;

    impl = "/**\n * " + concept.getComments() + "\n */\n";
    impl += "public ";
    if (form.startsWith("Abstract"))
      impl += "abstract ";
    impl += "class " + concept.getName() + " {\n" +
            getImplementation(concept, form, true) + "}";
    setImpl(concept, impl);
    if (log != null)
      log.log("Implemented " + concept.getName() + " as " + form);
  } //conceptAsClass

  /**
   * Implements the concept as an Interface.
   *
   * @param concept The concept to implement.
   */
  protected void conceptAsInterface(Concept concept)
  {
    String impl;

    impl = "/**\n * " + concept.getComments() + "\n */\n";
    impl += "public interface " + concept.getName() + " {\n" +
           getImplementation(concept, "Interface", true) + "}";
    setImpl(concept, impl);
    if (log != null)
      log.log("Implemented " + concept.getName() + " as Interface");
  } //conceptAsInterface

  /**
   * Implements the concept as a Method.
   *
   * @param concept The concept to implement.
   * @param form The exact concept form.
   */
  protected void conceptAsMethod(Concept concept, String form)
  {
    String impl = new String();

    impl = "/**\n * " + concept.getComments() + "\n */\n";
    // get method form
    if (form.startsWith("Public"))
      impl += "public ";
    else if (form.startsWith("Private"))
      impl += "private ";
    else if (form.startsWith("Abstract"))
      impl += "public abstract ";
    else if (form.startsWith("Static"))
      impl += "public static ";
    // append method
    impl += "void " + concept.getName() + "()";
    if (form.startsWith("Abstract"))
      impl += ";";
    else
     impl += " {\n" + getImplementation(concept, form, true) + "}";
    setImpl(concept, impl);
    if (log != null)
      log.log("Implemented " + concept.getName() + " as " + form);
  } //conceptAsMethod

  /**
   * Implements the concept as a ConstructorMethod.
   * @param concept The concept to implement.
   * @param form The form to implement the concept as.
   */
  protected void conceptAsConstructorMethod(Concept concept, String form)
  {
    String impl = new String();

    impl = "/**\n * " + concept.getComments() + "\n */\n";
    if (form.startsWith("Public"))
      impl += "public ";
    else if (form.startsWith("Private"))
      impl += "private ";
    impl += "<?name?>()";
    impl += " {\n" + getImplementation(concept, form, true) + "}";
    setImpl(concept, impl);
    if (log != null)
      log.log("Implemented " + concept.getName() + " as " + form);
  } //conceptAsConstructorMethod

  /**
   * Implements the concept as another form.
   * @param concept The concept to implement.
   * @param form The form to implement the concept as.
   */
  protected void conceptAsOther(Concept concept, String form)
  {
    String impl;

    impl = getImplementation(concept, form, false);
    setImpl(concept, impl);
    if (log != null)
      log.log("Implemented " + concept.getName() + " as " + form);
  } //conceptAsOther

  /**
   * Weaves the designated implementations into the pointers.
   *
   * @param model The source model to read the pointers from.
   * @throws CoComposeException if weaving failed.
   */
  protected void weavePointers(Model model) throws CoComposeException
  {
    Hashtable implementations = getImpls();
    Enumeration entities = implementations.keys();
    Object entity;

    while (entities.hasMoreElements())
    { // no unique realms used in Java, so all entities are Concepts.
      entity = entities.nextElement();
      weavePointers(model, entity);
    } //while
  } //weavePointers

  /**
   * Weaves the designated implementations into the pointers for the given
   * concept.
   *
   * @param model The source model to read the pointers from.
   * @param entity The entity of which to adapt the implementation.
   * @throws CoComposeException if weaving failed.
   */
  protected void weavePointers(Model model, Object entity)
    throws CoComposeException
  {
    Hashtable implementations = getImpls();
    String impl, subImpl, attr, result = new String();
    Object target;
    int ptrIndex = -1, endIndex, prevIndex = 0, ptr, attrIndex;

    if (log != null)
    {
      log.log("Weaving pointers for " + entity);
      log.incLogLevel();
    }
    try
    {
      // find implementation
      if (!implementations.containsKey(entity))
      {
        if (log != null)
          log.log("Error: no implementation found for entity");
        throw new CoComposeException("No implementation found for entity");
      }
      impl = (String) implementations.get(entity);
      // search and replace pointers
      while ((ptrIndex = impl.indexOf('<', ptrIndex)) != -1)
      { // pointer found
        ptrIndex++;
        endIndex = impl.indexOf('>', ptrIndex);
        // pointer attribute
        attrIndex = impl.indexOf('.', ptrIndex);
        if ((attrIndex > -1) && (attrIndex < endIndex))
          attr = impl.substring(attrIndex + 1, endIndex);
        else
          attr = null;
        try
        {
          if (attr == null)
            ptr = Integer.parseInt(impl.substring(ptrIndex, endIndex));
          else
            ptr = Integer.parseInt(impl.substring(ptrIndex, attrIndex));
        }
        catch (Exception e)
        { // this is no pointer, so continue
          continue;
        }
        // we have a pointer, so retrieve corresponding implementation
        if (log != null)
          log.log("Pointer " + ptr + " found");
        target = model.getNodeFromId(ptr);
        weavePointers(model, target); // weave sub entity first
        subImpl = getImplPart(attr, (String) implementations.get(target));
        result += impl.substring(prevIndex, ptrIndex - 1) + indent(subImpl).trim();
        prevIndex = endIndex + 1;
      } //while
      result += impl.substring(prevIndex);
      // put adapted implementation
      implementations.put(entity, result);
    }
    finally
    {
      if (log != null)
        log.decLogLevel();
    }
  } //weavePointers

  /**
   * Substitutes predefined variables. Known variables: <?name?>
   * @throws CoComposeException if substituting failed.
   */
  protected void substVars()
    throws CoComposeException
  {
    Hashtable implementations = getImpls();
    Enumeration entities = implementations.keys();
    Object entity;
    String impl;
    int varIndex = -1;

    if (log != null)
    {
      log.log("Substituting variables");
      log.incLogLevel();
    }
    try
    {
	  while (entities.hasMoreElements())
      {
        entity = entities.nextElement();
        impl = (String) implementations.get(entity);
        impl = substVars(entity, impl);
        implementations.put(entity, impl);
      }
    }
    finally
    {
      if (log != null)
        log.decLogLevel();
    }
  } //substVars

  /**
   * Retrieves the right part from the implementation.
   *
   * @param attr The pointer attributes, if any.
   * @param impl The current implementation.
   * @return The correct part of the implementation.
   * @throws CoComposeException if the part could not be found.
   */
  protected String getImplPart(String attr, String impl)
    throws CoComposeException
  {
    if (attr != null)
    {
      if (log != null)
        log.log("Pointer attribute found");
      return getPointerAttribute(attr, impl);
    }
    else
    {
      return impl;
    }
  } //getImplPart

  /**
   * Gets the pointer attribute part of the implementation for the given
   * entity.
   *
   * @param attr The pointer attribute identifier.
   * @param impl The implementation to get the attribute part for.
   * @throws CoComposeException if the part could not be found.
   */
  protected String getPointerAttribute(String attr, String impl)
    throws CoComposeException
  {
    int endIndex = 0, startIndex;
    String imports;

    // Dispatch attribute
    if (attr.equals("header"))
    { // get Java method header
      if (log != null)
        log.log("Header attribute found");
      try
      { // skip and save any import statements
        while ((startIndex = impl.indexOf("import", endIndex)) != -1)
        {
          endIndex = impl.indexOf(';', startIndex) + 1;
        }
        imports = impl.substring(0, endIndex);
        // skip comments
        startIndex = Math.max(impl.indexOf("*/") + 1, endIndex);
        // select only return type and method name
        if ((endIndex = impl.indexOf('{', startIndex)) == -1)
          endIndex = impl.indexOf(';', startIndex);
        startIndex = impl.lastIndexOf('(', endIndex);
        startIndex = impl.lastIndexOf(' ', startIndex);
        // check if return type present
        if (impl.lastIndexOf(' ', startIndex - 1) != -1)
          startIndex = impl.lastIndexOf(' ', startIndex - 1);
        impl = impl.substring(startIndex, endIndex) + imports;
      } //try
      catch (StringIndexOutOfBoundsException se)
      { // if this happens, there is a bug
        if (log != null)
          log.log("Error in getPointerAttribute: " + se.getMessage() +
            ", impl = \"" + impl + "\"");
        se.printStackTrace();
        throw new CoComposeException("Error in getPointerAttribute: " +
          se.getMessage() + ", impl = \"" + impl + "\"");
      } //catch
    }
    return impl;
  } //getPointerAttribute

  /**
   * Moves the import declarations in the implementations to the right place.
   */
  protected void moveImports()
  {
    Hashtable implementations = getImpls();
    Enumeration entities = implementations.keys();
    Object entity;
    String impl, imports;
    int startIndex, importIndex, endIndex, count;

    if (log != null)
    {
      log.log("Moving import declarations");
      log.incLogLevel();
    }
    try
    {
      while (entities.hasMoreElements())
      {
        entity = entities.nextElement();
        impl = ((String) implementations.get(entity)).trim();
        if ((startIndex = impl.indexOf("package ")) != -1)
          startIndex = impl.indexOf(';', startIndex) + 1;
        else
          startIndex = 0;
        // find class or interface start, or continue
//        if ((startIndex = impl.indexOf("class")) == -1)
//          if ((startIndex = impl.indexOf("interface")) == -1)
//            continue;
//        if ((endIndex = impl.lastIndexOf('\n', startIndex)) != -1)
//          startIndex = endIndex + 1;
//        if ((endIndex = impl.lastIndexOf("/*", startIndex)) != -1)
//          startIndex = endIndex;
        // find import declarations and move them
        count = 0;
        imports = new String();
        while ((importIndex = impl.indexOf("import ", startIndex)) != -1)
        { // find and copy import declaration
          endIndex = impl.indexOf(";", importIndex) + 1;
          imports += impl.substring(importIndex, endIndex) + '\n';
          // remove import declaration
          impl = impl.substring(0, importIndex).trim() + impl.substring(endIndex);
          count++;
        } //while importIndex
        if (count > 0)
        { // insert import declarations and set updated implementation
          impl = impl.substring(0, startIndex) + '\n' + imports +
            impl.substring(startIndex);
          if (log != null)
            log.log("Moved " + count + " import declarations " +
              entity);
          implementations.put(entity, impl);
        }
      } //while impls
    }
    finally
    {
      if (log != null)
        log.decLogLevel();
    }
  } //moveImports

  /**
   * Removes superfluous newlines in the implementations.
   */
  protected void removeExtraNewlines()
  {
    Hashtable implementations = getImpls();
    Enumeration entities = implementations.keys();
    Object entity;
    String impl, imports;

    int startIndex, importIndex, endIndex, count;

    if (log != null)
    {
      log.log("Removing superfluous newlines");
      log.incLogLevel();
    }
    try
    {
      while (entities.hasMoreElements())
      {
        entity = entities.nextElement();
        impl = ((String) implementations.get(entity)).trim();
        while ((startIndex = impl.indexOf("\n  \n")) != -1)
          impl = impl.substring(0, startIndex) + impl.substring(startIndex + 3);
        while ((startIndex = impl.indexOf("\n\n")) != -1)
          impl = impl.substring(0, startIndex) + impl.substring(startIndex + 1);
        implementations.put(entity, impl);
      } //while impls
    }
    finally
    {
      if (log != null)
        log.decLogLevel();
    }
  } //removeExtraNewlines

  /**
   * Writes the implementation of the model.
   *
   * @param model The model to write the implementation for.
   * @param outdir The directory to save the implementation files.
   * @throws CoComposeException if writing the implementation failed.
   */
  protected void writeImplementation(Model model, String outdir)
    throws CoComposeException
  {
    Hashtable implementations = getImpls();
    Enumeration entities = implementations.keys();
    Concept concept;
    String impl;

    if (log != null)
    {
      log.log("Writing implementations to files");
      log.incLogLevel();
    }
    try
    {
      while (entities.hasMoreElements())
      { // no unique realms used in Java, so all entities are Concepts.
        concept = (Concept) entities.nextElement();
        if (concept.getForm().endsWith("Class") ||
            concept.getForm().equals("Interface"))
        { // only write classes and interfaces to files
          impl = (String) implementations.get(concept);
          writeFile(outdir, concept.getName() + ".java", impl);
          if (log != null)
            log.log(concept.getName() + " written");
        }
      } //while
    }
    finally
    {
      if (log != null)
        log.decLogLevel();
    }
  } //writeImplementation

  //////////////////////////////////////////////////////////////////////////////
  // Public methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Returns the target language this GenerateImplementationVisitor supports.
   *
   * @return The target language.
   */
  public String getTargetLanguage()
  {
    return "Java";
  } //getTargetLanguage

  /**
   * Visits a Model object.
   *
   * @param model The model to visit.
   * @throws CoComposeException if visiting operation failed.
   */
  public void visitModel(Model model) throws CoComposeException
  {
    Concept concept;
    Composite composite;
    String outdir = OUTPUT_DIR + getTargetLanguage() + File.separator +
      model.getName();

    if (log != null)
    {
      log.log("-- Started generating implementation for " + model.getName());
      log.incLogLevel();
    }
    try
    {
      // generate implementation
      super.visitModel(model);
      // dump implementation in .dmp file
      dumpImplementation(model, outdir);
      // weave method pointers
      weavePointers(model);
      // move import declarations
      moveImports();
      // substitute predefined variables
      substVars();
      // remove superfluous newlines
      removeExtraNewlines();
      // write implementation
      writeImplementation(model, outdir);
    }
    finally
    {
      if (log != null)
      {
        log.decLogLevel();
        log.log("-- Completed generating implementation for " + model.getName());
      }
    }
  } //visitModel

  /**
   * Generates implementation for a Concept object.
   *
   * @param concept The concept to visit.
   * @throws CoComposeException if generating implementation failed.
   */
  public void visitConcept(Concept concept) throws CoComposeException
  {
    String form = concept.getForm();

    if (log != null)
    {
      log.log("Generating implementation for concept " + concept.getName());
      log.incLogLevel();
    }
    try
    {
      super.visitConcept(concept);
      // now dispatch to specific code generators
      if (form.equals(J_CLASS))
        conceptAsClass(concept, form);
      else if (form.equals(J_ABSCLASS))
        conceptAsClass(concept, form);
      else if (form.equals(J_INTERFACE))
        conceptAsInterface(concept);
      else if (form.equals(J_PUBCONSTR))
        conceptAsConstructorMethod(concept, form);
      else if (form.equals(J_PRICONSTR))
        conceptAsConstructorMethod(concept, form);
      else if (form.equals(J_PUBMETHOD))
        conceptAsMethod(concept, form);
      else if (form.equals(J_PRIMETHOD))
        conceptAsMethod(concept, form);
      else if (form.equals(J_ABSMETHOD))
        conceptAsMethod(concept, form);
      else if (form.equals(J_STAMETHOD))
        conceptAsMethod(concept, form);
      else
        conceptAsOther(concept, form);
    }
    finally
    {
      if (log != null)
        log.decLogLevel();
    }
  } //visitConcept
} //class JavaImplementationVisitor
