/**
 * Title:        CoCompose
 * Copyright:    Copyright (c) 2002, 2003
 * Author:       Dennis Wagelaar
 * Organisation: University of Twente, Vrije Universiteit Brussel
 * Description:  CoCompose Design Language Tool
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License, version 2 as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  (See the file "COPYING" that is included with this source distribution.)
 */

package translate;

import java.util.*;
import EDU.utwente.CoCompose.lang.*;
import EDU.utwente.CoCompose.util.*;

/**
 * Generates SEESCOA component language implementation for a CoCompose model.
 *
 * @author Dennis Wagelaar
 * @version 0.1.15, 22 October 2003
 */

public class SEESCOAImplementationVisitor extends JavaImplementationVisitor
{
  //////////////////////////////////////////////////////////////////////////////
  // Constants
  //////////////////////////////////////////////////////////////////////////////

  private static final String S_COMPONENTCLASS = "ComponentClass";
  private static final String S_PORT           = "Port";
  private static final String S_MESSAGE        = "Message";

  //////////////////////////////////////////////////////////////////////////////
  // Constructors
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Constructs a SEESCOAImplementationVisitor.
   */
  public SEESCOAImplementationVisitor()
  {
  } //SEESCOAImplementationVisitor

  //////////////////////////////////////////////////////////////////////////////
  // Protected methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Implements the concept as a Componentclass.
   *
   * @param concept The concept to implement.
   * @param form The exact concept form.
   */
  protected void conceptAsComponentClass(Concept concept, String form)
  {
    String impl;

    impl = "/**\n * " + concept.getComments() + "\n */\n";
    impl += "componentclass ";
    impl += concept.getName() + " {\n" +
            getImplementation(concept, form, true) + "}";
    setImpl(concept, impl);
    if (log != null)
      log.log("Implemented " + concept.getName() + " as " + form);
  } //conceptAsComponentClass

  /**
   * Implements the concept as a Message.
   *
   * @param concept The concept to implement.
   * @param form The exact concept form.
   */
  protected void conceptAsMessage(Concept concept, String form)
  {
    String impl = new String();

    impl = "/**\n * " + concept.getComments() + "\n */\n";
    impl += "message ";
    impl += concept.getName() + "()";
    if (form.startsWith("Abstract"))
      impl += ";";
    else
     impl += " {\n" + getImplementation(concept, form, true) + "}";
    setImpl(concept, impl);
    if (log != null)
      log.log("Implemented " + concept.getName() + " as " + form);
  } //conceptAsMessage

  /**
   * Writes the implementation of the model.
   *
   * @param model The model to write the implementation for.
   * @param outdir The directory to save the implementation files.
   * @throws CoComposeException if writing the implementation failed.
   */
  protected void writeImplementation(Model model, String outdir)
    throws CoComposeException
  {
    Hashtable implementations = getImpls();
    Enumeration entities = implementations.keys();
    Concept concept;
    String impl, form;

    if (log != null)
    {
      log.log("Writing implementations to files");
      log.incLogLevel();
    }
    try
    {
      while (entities.hasMoreElements())
      { // no unique realms used in SEESCOA, so all entities are Concepts.
        concept = (Concept) entities.nextElement();
        form = concept.getForm();
        if (form.equals("AbstractClass") || form.equals("Class") ||
            form.equals("Interface"))
        { // write classes and interfaces to files
          impl = (String) implementations.get(concept);
          writeFile(outdir, concept.getName() + ".java", impl);
          if (log != null)
            log.log(concept.getName() + " written");
        }
        else if (concept.getForm().equals("ComponentClass"))
        { // write componentclasses to files
          impl = (String) implementations.get(concept);
          writeFile(outdir, concept.getName() + ".component", impl);
          if (log != null)
            log.log(concept.getName() + " written");
        }
        else if (concept.getForm().endsWith("Port"))
        { // write componentclasses to files
          impl = (String) implementations.get(concept);
          writeFile(outdir, concept.getName() + ".port", impl);
          if (log != null)
            log.log(concept.getName() + " written");
        }
      } //while
    }
    finally
    {
      log.decLogLevel();
    }
  } //writeImplementation

  //////////////////////////////////////////////////////////////////////////////
  // Public methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Returns the target language this GenerateImplementationVisitor supports.
   *
   * @return The target language.
   */
  public String getTargetLanguage()
  {
    return "SEESCOA";
  } //getTargetLanguage

  /**
   * Generates implementation for a Concept object.
   *
   * @param concept The concept to visit.
   * @throws CoComposeException if generating implementation failed.
   */
  public void visitConcept(Concept concept) throws CoComposeException
  {
    String form = concept.getForm();

    if (log != null)
    {
      log.log("Generating implementation for concept " + concept.getName());
      log.incLogLevel();
    }
    try
    {
      // now dispatch to specific code generators
      if (form.equals(S_COMPONENTCLASS))
        conceptAsComponentClass(concept, form);
      else if (form.endsWith(S_PORT))
        conceptAsClass(concept, form);
      else if (form.equals(S_MESSAGE))
        conceptAsMessage(concept, form);
      else
        super.visitConcept(concept);
    }
    finally
    {
      if (log != null)
        log.decLogLevel();
    }
  } //visitConcept
} //class SEESCOAImplementationVisitor